﻿namespace Invadurz

open System
open System.Windows
open System.Windows.Controls
open System.Windows.Data
open System.Windows.Input
open System.Windows.Media
open System.Windows.Media.Imaging
open System.Windows.Shapes
open System.Windows.Threading

[<AutoOpen>]
module Resources =
    let settings = System.IO.IsolatedStorage.IsolatedStorageSettings.ApplicationSettings
    let toGradientStops stops =
        let collection = GradientStopCollection()
        stops
        |> List.map (fun (color,offset) -> GradientStop(Color=color,Offset=offset)) 
        |> List.iter collection.Add
        collection
    let skyBrush = 
        let darkBlue = Color.FromArgb(255uy,0uy,0uy,64uy)
        let stops = [Colors.Black,0.0; darkBlue,1.0] |> toGradientStops
        LinearGradientBrush(stops, 90.0)
    let toArgb (color:Color) = 
        int color.A <<< 24 ||| 
        int color.R <<< 16 ||| 
        int color.G <<< 8  ||| 
        int color.B
    let whiteBrush = SolidColorBrush Colors.White
    let greenBrush = SolidColorBrush Colors.Green
    let createTextBlock () =
            TextBlock(
                Foreground=whiteBrush, 
                FontSize=15.0, 
                FontWeight=FontWeights.ExtraBold,
                Effect=Effects.BlurEffect(Radius=10.0/3.0))
    let wait n = seq { for x = 1 to n do yield () }

[<AutoOpen>]
module Invaders =
    let toImage (bitmap:#BitmapSource) =
        let w = bitmap.GetValue(BitmapSource.PixelWidthProperty) :?> int
        let h = bitmap.GetValue(BitmapSource.PixelHeightProperty) :?> int
        Image(Source=bitmap,Stretch=Stretch.Fill,Width=float w*2.0,Height=float h*2.0) 
    let createRow (invader:(int * int[]) [],i) (x',y') =
        let bitmaps = invader |> Array.map toBitmap
        [|0..11|]
        |> Array.map (fun x ->
            let images = bitmaps |> Array.map toImage
            images |> Seq.iter (fun image -> image.Visibility <- Visibility.Collapsed)
            images.[i].Visibility <- Visibility.Visible
            let animation = seq {
                let i = ref i
                while true do
                    yield! wait 6
                    images.[!i].Visibility <- Visibility.Collapsed
                    i := (!i+1) % images.Length
                    images.[!i].Visibility <- Visibility.Visible
            }
            Sprite (images |> Seq.cast, x' + 32.0*float x,y',animation)
            //let image = Image(Width=32.0,Height=24.0,Source = BitmapImage(Uri("hamster.png",UriKind.RelativeOrAbsolute)))
            //Sprite(image,x' + 32.0*float x,y')
        )
    let createInvaders () =
        [|  invader1,0
            invader2,0
            invader2,1
            invader3,0
            invader3,1 |]
        |> Array.mapi (fun row invader -> (row,invader))
        |> Array.collect (fun (row,invader) -> createRow invader (0.0, 24.0*float row))

type Bunker (x,y) =
    let toSpans (width,xs:int []) =
        let mins = Array.create width xs.Length
        let maxs = Array.create width -1
        xs |> Array.iteri (fun y xs ->
            for x = 0 to width-1 do
                let bit = 1 <<< (width - 1 - x) 
                if xs &&& bit = bit then
                    mins.[x] <- min y mins.[x]
                    maxs.[x] <- max y maxs.[x]
        )
        Array.zip mins maxs
    let spans = bunker |> toSpans
    do spans |> Seq.iter (fun (y,y') -> System.Diagnostics.Debug.WriteLine(sprintf "%d %d" y y') )
    let bitmap = bunker |> toBitmap
    let width, height = bunker |> fst, bunker |> snd |> Seq.length
    let image = bitmap |> toImage
    do  image.Opacity <- 0.90
    let sprite = Sprite(image,x,y)
    member this.Sprite = sprite
    member this.Busted (x',y',dy) =
        if sprite.HitTest(x',y') then
            let x = (x' - x) / 2.0 |> int
            let y = (y' - y) / 2.0 |> int
            let s = sign dy |> int
            let zt,zb = spans.[x]
            if zt <= zb then
                let z = if s > 0 then zt else zb
                for i = 0 to 3 do
                    let k = z + i * s
                    if k >= 0 && k < height then
                        bitmap.Pixels.[x + k * width] <- 0
                spans.[x] <- 
                    if s > 0 then (zt+4,zb) else (zt,zb-4)
                true
            else false
        else false

type Action = Left | Right | Fire

type GameControl () as control =
    inherit UserControl ()

    let mutable disposables = []
    let remember disposable = disposables <- disposable :: disposables
    let dispose (d:IDisposable) = d.Dispose()
    let forget () = disposables |> List.iter dispose; disposables <- []

    let width, height = 512.0, 384.0
    do  control.Width <- width; control.Height <- height

    let toAction = function
        | Key.Z -> Some Left
        | Key.X -> Some Right
        | Key.Space -> Some Fire
        | _ -> None

    let mutable actions = set []

    let startObservingKeys () =
        control.KeyDown
        |> Observable.choose (fun ke -> toAction ke.Key)
        |> Observable.subscribe(fun action -> 
            actions <- Set.add action actions)
        |> remember
        control.KeyUp
        |> Observable.choose (fun ke -> toAction ke.Key)
        |> Observable.subscribe (fun action -> 
            actions <- Set.remove action actions)
        |> remember

    let screen = Canvas(Background=skyBrush)
    let add (x:#UIElement) = screen.Children.Add x
    let remove (x:#UIElement) = screen.Children.Remove x |> ignore

    let mutable score = 0
    let scoreText = createTextBlock()
    do  scoreText.Text <- sprintf "SCORE  %d" score
    do  add scoreText

    let mutable highScore = 
        if settings.Contains "HighScore" then settings.["HighScore"] :?> int else 0
    let highScoreText = createTextBlock()
    do  highScoreText.Text <- sprintf "HIGH  %d" highScore
    do  highScoreText.Width <- width
    do  highScoreText.TextAlignment <- TextAlignment.Center
    do  add highScoreText

    let mutable lives = 3
    let livesText = createTextBlock()
    do  livesText.Text <- sprintf "LIVES  %d" lives
    do  livesText.Width <- width
    do  livesText.TextAlignment <- TextAlignment.Right
    do  add livesText

    let mutable sheetX, sheetY = 64.0, 24.0
    let mutable direction = 1.0
    let mutable invaders = [||]

    let mutable sheet = Sprite(invaders |> Seq.cast,sheetX,sheetY)

    let createSheet () = 
        invaders <- createInvaders ()
        sheetX <- 64.0
        sheetY <- 32.0
        sheet <- Sprite(invaders |> Seq.cast,sheetX,sheetY)

    let cleanSheet () =
        direction <- 1.0
        createSheet()
        add sheet

    let newSheet () = 
        direction <- direction + 0.1
        remove sheet
        createSheet()
        add sheet

    do cleanSheet ()

    let bunkers =
        [1..4] 
        |> List.map (fun x -> Bunker((float x*width/5.0)-8.0,height-64.0))
    do  bunkers |> List.map (fun bunker -> bunker.Sprite) |> List.iter add

    let cannonWidth = 24.0
    let mutable cannonX, cannonY = 100.0, height-32.0
    let mutable cannonReload = 0
    let cannon = Sprite(laserCannon |> toBitmap |> toImage, cannonX, cannonY)

    let mutable missiles = []
    let mutable bombs = []
    let mutable particles = []

    let rand = Random()

    let explode (x,y) =
        let boom (dx,dy) = 
            let line = Line(X1=dx,Y1=dy,X2=0.0,Y2=0.0,Stroke=whiteBrush)
            let sprite = Sprite(line,x+dx,y+dy)
            add sprite
            particles <- (sprite,x,y,dx,dy,100) :: particles
        [1..100] 
        |> Seq.map (fun _ ->
            let r = rand.NextDouble() * 2.0 * Math.PI
            let v = 4.0 *  rand.NextDouble() + 2.0
            v * cos r, v * sin r 
        )
        |> Seq.iter boom

    let updateCannon () =
        if actions.Contains Left && cannonX > 0.0 then cannonX <- cannonX - 2.0
        if actions.Contains Right && cannonX < width - cannonWidth then cannonX <- cannonX + 2.0
        cannon.MoveTo(cannonX,cannonY)
        if cannonReload > 0 then cannonReload <- cannonReload - 1
        if actions.Contains Fire && cannonReload = 0 then 
            let rectangle = Rectangle(Width=2.0,Height=4.0, Fill=whiteBrush)
            let missile = Sprite(rectangle, cannonX, cannonY-4.0)
            add missile 
            missiles <- 
                (missile,cannonX+10.0,cannonY,-4.0) :: missiles
            cannonReload <- 8

    let updateInvaders () =
        let startX, endX, count = 
            invaders |> Seq.fold (fun (lo,hi,count) invader ->
                let x = invader.X
                min x lo, max x hi, count+1
            ) (Double.MaxValue, Double.MinValue,0)
        if direction > 0.0 then (sheetX + direction) > (width - (endX+24.0))
        else sheetX + startX - direction < 0.0
        |> (function
            | true -> sheetY <- sheetY + 8.0; direction <- -direction
            | false -> sheetX <- sheetX + (((float (60 - count))/15.0 + 1.0) * direction)
        )

        sheet.MoveTo(sheetX,sheetY)
        invaders |> Seq.iter (fun invader -> invader.Animate())

        invaders |> Seq.iter (fun invader ->
            if rand.Next(count * 16) = 0 then
                let rectangle = Rectangle(Width=2.0,Height=4.0, Fill=whiteBrush)
                let lines = 
                    [0.0,0.0,2.0,2.0;2.0,2.0,0.0,4.0;0.0,4.0,2.0,6.0]
                    |> List.map (fun (x1,y1,x2,y2) -> Line(X1=x1,Y1=y1,X2=x2,Y2=y2,Stroke=whiteBrush))
                let bomb = Sprite(lines |> Seq.cast,0.0,0.0)
                add bomb 
                bombs <- (bomb,sheetX + invader.X+12.0,sheetY + invader.Y+12.0,3.0) :: bombs
        )

    let updateWeapons weapons =
        let current, expired =
            weapons
            |> List.map (fun (m,x,y,dy) -> (m,x,y+dy,dy))
            |> List.partition (fun (_,x,y,dy) -> 
                y>0.0 && y<height &&
                bunkers |> List.exists (fun bunker -> 
                    let y = if dy > 0.0 then y+6.0 else y
                    bunker.Busted(x+1.0,y,dy)
                    ) |> not
            )
        expired |> List.iter (fun (weapon,_,_,_) -> remove weapon)
        current |> List.iter (fun (weapon:Sprite,x,y,_) -> weapon.MoveTo(x,y))
        current

    let updateScore () =
        let missilesHit, invadersHit =
            invaders |> Seq.fold (fun acc invader ->
                missiles |> Seq.tryFind (fun (_,x,y,_) ->
                    invader.HitTest(x - sheetX, y - sheetY)
                )
                |> (function 
                    | Some (missile,_,_,_) -> (missile,invader)::acc 
                    | None -> acc)
            ) []
            |> List.unzip
        invadersHit |> List.iter (fun invader -> sheet.Children.Remove invader |> ignore)
        invaders <- invaders |> Array.filter (fun invader -> invadersHit |> List.exists((=) invader) |> not)
        missilesHit |> List.iter (fun missile -> explode(missile.X,missile.Y); remove missile)
        missiles <- missiles |> List.filter (fun (missile,_,_,_) -> missilesHit |> List.exists((=) missile) |> not)
        if invadersHit.Length > 0 then
            score <- score + invadersHit.Length * 10
            scoreText.Text <- sprintf "SCORE  %d" score

    let updateParticles () =
        let alive, dead =
            particles 
            |> List.map (fun (p,x,y,dx,dy,count) -> p,x+dx,y+dy,dx,dy,count-1)
            |> List.partition (fun (p,x,y,dx,dy,count) -> count>0)
        dead |> List.iter (fun (p,x,y,dx,dy,count) -> remove p)
        particles <- alive
        particles |> List.iter (fun (p,x,y,_,_,_) -> p.MoveTo(x,y))

    let updateScreen () =
        updateInvaders ()
        missiles <- updateWeapons missiles 
        bombs <- updateWeapons bombs
        updateParticles()

    let play () =
        updateCannon ()
        updateScreen ()
        updateScore()

    let isCannonHit () =
        bombs |> Seq.exists (fun (bomb,x,y,_) -> cannon.HitTest(x+1.0,y+6.0))

    let layout = Grid()
    do  layout.Children.Add screen
    do  control.Content <- layout

    let message s =
        let t = createTextBlock()
        t.Text <- s
        t.HorizontalAlignment <- HorizontalAlignment.Center
        t.VerticalAlignment <- VerticalAlignment.Center
        t

    let updateLoop () = seq {
        while true do
        add cannon
        while lives > 0 do
            cannon.Opacity <- 0.5
            for i = 1 to 25 do play() |> ignore; yield () done
            cannon.Opacity <- 1.0
            while (play(); not (isCannonHit()) && invaders.Length>0) do yield ()
            if isCannonHit() then 
                lives <- lives - 1
                livesText.Text <- sprintf "LIVES  %d" lives
                remove cannon
                for i = 1 to 25 do updateScreen(); yield () done
                if lives > 0 then add cannon
            elif invaders.Length = 0 then
                let mess = message "COMPLETED"
                layout.Children.Add mess
                for i = 1 to 100 do updateScreen(); yield() done
                layout.Children.Remove mess |> ignore
                newSheet()
        done
        if score > highScore then
            highScore <- score
            settings.["HighScore"] <- highScore
            highScoreText.Text <- sprintf "HIGH  %d" highScore
        let mess = message "GAME OVER"
        layout.Children.Add mess
        for i = 1 to 100 do updateScreen(); yield() done
        layout.Children.Remove mess |> ignore
        remove sheet
        cleanSheet ()
        lives <- 3
        score <- 0
        livesText.Text <- sprintf "LIVES  %d" lives
        scoreText.Text <- sprintf "SCORE  %d" score
        done
        }

    let startGame () =
        let e = (updateLoop ()).GetEnumerator()
        let timer = DispatcherTimer()
        timer.Interval <- TimeSpan.FromMilliseconds(20.0)
        timer.Tick
        |> Observable.subscribe (fun _ -> e.MoveNext() |> ignore)
        |> remember
        timer.Start()

    do  let t = message "Click to Start"
        layout.Children.Add t
        control.MouseLeftButtonUp
        |> Observable.subscribe (fun _ -> 
            forget()
            startObservingKeys ()
            layout.Children.Remove t |> ignore
            startGame() 
        )
        |> remember

    interface System.IDisposable with
        member this.Dispose() = forget()